/* 
 *  $Id: Capsule.java,v 1.2 2000/04/02 21:30:37 andrew Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import utils.*;


/**
 * Base class for all types of capsules and their processing
 *
 * @author David Wetherall
 */


public abstract class Capsule {
  protected abstract byte[] mid();
  protected abstract byte[] pid();

  // added for djw fix
  protected abstract byte[] gid();

  public TypeID getCapsuleID() {
    return new TypeID(mid());
  }


  /** All capsules originate from some source node. */
  int src; // accessed directly by channel
  private int dst;

  /** Previous active node, required for code loading protocol */
  int previous; // accessed directly by channel

  public int getPrevious() {
    return previous;
  }

  int resources; // accessed directly by channel and nodecache

  public int getResources() {
    return resources;
  }
  
  Xdr xdr = null;
  boolean saveBuffer = false;

  final static protected int LVL = Entity.L[1]; 

  // set by channel during reception
  Channel chan;

  // for access by the capsule code
  public ChannelObject getChannel() {
    return (chan != null) ? new ChannelObject(chan) : null;
  }


  /** Access point to initialze capsule methodIDs */
  public static byte[] findMID(String name) {
    // redirect access point to the real implementation
    return Method.findMID(name);
  }

  /** Access point to initialze capsule groupIDs */
  public static byte[] findGID(String name) {
    // redirect access point to the real implementation
    return Method.findGID(name);
  }

  /** Access point to initialze capsule protocolIDs */
  public static byte[] findPID(String name) {
    // redirect access point to the real implementation
    return Method.findPID(name);
  }

  /** There is no longer code to find a named extension; these methods
      are in the node class  */
  
  abstract public boolean evaluate(Node n);
  
  void setSrc (int source) {
    this.src = source;
  }

  public int getSrc() {
    return src;
  }

  public int getDst() {
    return dst;
  }

  public void setDst(int dst) {
    this.dst = dst;
  }

  public int length() {
    return Xdr.TYPEID + Xdr.INT + Xdr.INT + Xdr.INT + Xdr.INT;
  }
  
  public Xdr encode() {
    Xdr xdr = new Xdr(this.xdr, this.xdr.index);
    xdr.index = 0;
    xdr.PUT(this.mid(), 0, Xdr.TYPEID);
    xdr.PUT(src);
    xdr.PUT(dst);
    xdr.PUT(previous);
    xdr.PUT(resources);

    return xdr;
  }

  public Xdr decode() {
    Xdr xdr = new Xdr(this.xdr, this.xdr.index);
    
    // skip over the MID field
    // xdr.index += Xdr.TYPEID;
    // PANAMA line below added to access mid for debugging
    TypeID mid = xdr.TYPEID();

    src = xdr.INT();
    dst = xdr.INT();
    previous = xdr.INT();
    resources = xdr.INT();

    return xdr;
  }

  /** For bootstrapping capsule code loading */
  final static int peekPrevious(Xdr xdr) {
    xdr = new Xdr(xdr, xdr.index + Xdr.TYPEID + Xdr.INT + Xdr.INT);
    return xdr.INT();
  }

  /** Prime a capsule created within the network for transmission */
  public final void prime(Capsule parent) {
    this.src = parent.src;
    int rnew = parent.resources--;
    if (rnew <= 0) throw new ResourceLimitException();
    this.resources = rnew;
  }

  /** Privledged local prime */
  final void prime(int src, int resources) {
    this.src = src;
    this.resources = resources;
  }

  public static Capsule create(Class cl, Xdr xdr) {
    Capsule c = null;
    try {
      c = (Capsule)(cl.newInstance());
    } 
    catch (Exception e) {
      Entity.error("Class.newInstance failed for "+ cl);
    }
    c.xdr = xdr;
    c.decode();
    return c;
  }

  public Capsule() { }
}




 
