/* 
 *  $Id: CodeCache.java,v 1.1.1.1 2000/03/13 21:26:49 andrew Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;
import java.util.Hashtable;


/**
 * Cache of Methods and CodeGroups held at a Node
 *
 * @author David Wetherall
 */


class CodeCache
  implements Runnable
  {
  private Method[] buckets;
  private Hashtable groups;
  private CodeGroup head;
  private CodeGroup tail;

  private int capacity;

  // number of hash buckets
  private int size;
  
  // current number of entries
  private int num;
  
  final static private boolean FLUSHCACHE = false;
  private Node owner;

  private void movetofront(CodeGroup group) {
    if (group == null || group.pin == true) return;

    if (head != group) {
      
      // we assume head is non-null since insertathead
      // should have been called for a non-null group
      Entity.assert(head != null, "operating on empty codegroup queue");

      CodeGroup n = group.next;
      CodeGroup p = group.prev;
      p.next = n;
      if (n != null) n.prev = p;
      if (tail == group) tail = p;
      head.prev = group;
      group.next = head;
      group.prev = null;
      head = group;
    }
  }


  private void insertathead(CodeGroup group) {
    
    // all calls are expected to be real
    Entity.assert(group != null, "called with null code group");
    
    if (group.pin == true) return;
    if (head != null) head.prev = group;
    group.next = head;
    group.prev = null;
    head = group;
    if (tail == null) tail = group;
  }

  
  private void delete(CodeGroup group) {

    // remove all method components from their buckets
    for (int i = 0; i < group.meths.size(); i++) {
      Method m = (Method)group.meths.elementAt(i);
      int hcode = m.hash;
      int bin = (hcode & 0x7FFFFFFF) % size;
      Method s = buckets[bin];
      Method prev = null;
      
      while (s != null) {
	if (s == m) {
	  if (prev != null) prev.next = s.next;
	  else buckets[bin] = s.next;
	  s.next = null;
	}
	prev = s; s = s.next;
	
	num--;
      }
    }
      
    // remove the group from the lru list and table
    CodeGroup n = group.next;
    CodeGroup p = group.prev;
    if (p != null) p.next = n;
    if (n != null) n.prev = p;
    if (tail == group) tail = p;
    if (head == group) head = n;
    group.next = group.prev = null;

    groups.remove(group.groupID);
  }
  
  
  public synchronized Method fastLookup(byte[] buf, int offset) {

    // this is taken from hashcode on typeid and must match
    int hcode = 
      ((buf[offset+0]+buf[offset+4] + buf[offset+8] + buf[offset+12]) << 24)+
      ((buf[offset+1]+buf[offset+5] + buf[offset+9] + buf[offset+13]) << 16)+
      ((buf[offset+2]+buf[offset+6] + buf[offset+10] + buf[offset+14]) << 8)+
      (buf[offset+3]+buf[offset+7] + buf[offset+11] + buf[offset+15]);
    int bin = (hcode & 0x7FFFFFFF) % size;
    
  outer:
    for (Method m = buckets[bin] ; m != null ; m = m.next) {
      if (m.hash == hcode) {
	byte[] bits = m.methodID.unwrap();
	
	for (int i = 0; i < 16; i++)
	  if (bits[i] != buf[i+offset]) continue outer;
	
	movetofront(m.group);
	return m;
      }
    }
    return null;
  }
  

  public synchronized Method lookup(TypeID key) {
    int hcode = key.hashCode();
    int bin = (hcode & 0x7FFFFFFF) % size;

    for (Method m = buckets[bin] ; m != null ; m = m.next) {
      if ((m.hash == hcode) && m.methodID.equals(key)) {
	movetofront(m.group);
	return m;
      }
    }
    return null;
  }


  public synchronized CodeGroup retrieve(TypeID key) {
    return (CodeGroup)groups.get(key);
  }


  public synchronized void enter(Method value) {
    TypeID key = value.methodID;
    int hcode = key.hashCode();
    int bin = (hcode & 0x7FFFFFFF) % size;

    for (Method m = buckets[bin] ; m != null ; m = m.next) {
      if ((m.hash == hcode) && m.methodID.equals(key)) {
       
	// not yet prepared to handle this
	Entity.error("CodeCache: method already in cache");
	return;
      }
    }
    
    // make a little room; should only fire once
    while (num >= capacity) delete(tail);
    
    value.next = buckets[bin];
    buckets[bin] = value;
    value.hash = hcode;
    num++;
    
    // and add to lru queue
    movetofront(value.group);
  }
  
  public synchronized void set(CodeGroup value) {
    TypeID key = value.groupID;

    if (groups.get(key) != null) {
      
      // not yet prepared to handle this
      Entity.error("CodeCache: codegroup already in cache");
      return;
    }

    groups.put(key, value);
    insertathead(value);
  }

  // run added for djw fix
  public void run()
  {
	while (true) 
	{
		try {
			Thread.sleep(1000);
		} catch (InterruptedException e) {;}

		synchronized (owner)
		{
			boolean changed = false;
			do
			{
				for (CodeGroup g = head; g != null; g = g.next)
				{
					if(!g.pin)
					{
						delete(g);
						changed = true;
						break;
					}
				}
			} while (changed);
		}
	}
  }


  // remove goes here!
  
  public CodeCache(Node owner, int capacity) {
    
    // arrange to expect only 4/3 entries per bucket
    this.capacity = capacity;
    this.size = (int)(capacity*0.75f + 0.5);
    this.num = 0;
 
    this.owner = owner;
    buckets = new Method[size];
    for (int i = 0; i < size; i++) buckets[i] = null;

    groups = new Hashtable();
    head = tail = null;

    if (FLUSHCACHE) new Thread(this).start();
  }

}



 
