/* 
 *  $Id: DLClassLoader.java,v 1.3 2000/04/02 21:30:37 andrew Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import java.util.Hashtable;

/**
 * Demand Loading ClassLoader.  Presents an interface to load a
 * code group from byte arrays received by the DL protocol or 
 * registered at the Node.  Heavily modified by Andrew Whitaker.
 *
 * @author David Wetherall
 */


// warning -- there be wicked interactions here.  this code is 
// heavily commented and asserted as a result. see debug also.

final class DLClassLoader extends ClassLoader {

  /** Determines whether capsules can breach the system load boundry. 
   *  Setting this to true effectively removes any capsule security checks. 
  */
  public static final boolean ALLOW_BREACH = false;

  /** Cache of system classes available to capsules  */
  private static Hashtable sharedRegistry = new Hashtable();
  
  /** Side information for loadClass -- the code group being loaded */
  private CodeGroup inProgress;

  /** Side information for loadClass -- local registration or DL protocol */
  private boolean isDL;
  
  /** Compile time flag for debugging statements */
  private static final boolean DEBUG = false;

  /** Debug print statement, redirects to logger */
  public void debug(String msg) {
    if (DEBUG) System.err.println("DLClassLoader: "+ msg);
  }

  /** A local helper routine */
  protected Class resolver (Class cl,boolean resolve) {
    if (resolve) resolveClass (cl);
    return cl;
  }

  /**
   * This method is the catch-all to intercept any unknown class 
   * references that are needed by capsules.  Eventually all classes
   * should be found already in the registry, or an error occurs.
   */

  protected synchronized Class loadClass(String name, boolean resolve) 
    throws ClassNotFoundException {
      
    if (inProgress != null) {
      debug("loadClass: called by codegroup for "+ name);
    } 
    else {
      debug("loadClass: called by system for "+ name);
    }
    
    // Check to see if the class has already been defined 
    Class cl = findLoadedClass(name);
    if (cl != null) {
      debug("loadClass: result found in registry for "+ name);
      return cl;
    }

    // Check the set of allowed system classes 
    cl = (Class) sharedRegistry.get(name);
    if (cl != null) {
      debug("loadClass: result found in shared registry for " + name);
      return cl;
    } 

    if ((inProgress != null) && isDL)  {
      // We have a demand (non-local) load on our hands 	
      CodeGroup g = inProgress;
      
      // first, find the right method
      for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);
	
	if (name.equals(m._cn)) {
	  debug("loadClass: calling defineClass for "+ name);

	  /* Really, the security manager should decide this with 
	   * the checkPackageDefinition call, but we don't necessarily
	   * have a security manager running.
	  */
	  if (name.startsWith("ants.") || name.startsWith("java."))
	    throw new SecurityException ("Illegal package definition");

	  cl = defineClass (name,m.code,0,m.code.length);
	  if (cl != null)
	    return resolver (cl,resolve);
	}
      }
    }
    else if ((inProgress != null) && (!isDL)) {
      // This is a local load 
      cl = Class.forName(name);
      if (cl != null)
	return cl;
    }
	 
    /* If we allow breached loads, check in the system classes */
    System.out.println ("Attempting breach:");
    if (ALLOW_BREACH) {
      cl = findSystemClass (name);
    }
    else throw new SecurityException("Attempted Breach Load");
    
    if (cl != null)
      return cl;
    else throw new ClassNotFoundException (name);
  }

  /**
   * Entry point used by a node to register code groups.
   * Loads ultimately come from the local filesystem
   */

  synchronized boolean loadForRegister(CodeGroup g) 
    throws Exception {

    Entity.assert(inProgress == null, "DLClassLoader: Interleaved load");
    
    debug("loadForRegister: begin loading "+ g);

    try {
      inProgress = g;
      isDL = false;
      
      for (int i = 0; i < g.meths.size(); i++)  {
	Method m = (Method)g.meths.elementAt(i);
	
	// load failures will throw an exception here, which
	// should percolate up to the application registering
	
	m.cl = loadClass(m._cn, true);
      }
    } 
    catch (Exception e) {
      throw e;
    } 
    finally {
      
      // must release
      inProgress = null;
    }

    debug("loadForRegister: end loading "+ g);
  
    return true;
  }


  /**
   * Entry point used by a the DL protocol load code groups.
   * Loads ultimately come from byte arrays.
   */

  synchronized boolean loadOnDemand(CodeGroup g) {
    Entity.assert(inProgress == null, "DLClassLoader: Interleaved load");
    
    debug("loadOnDemand: begin loading "+ g);

    boolean loaded = true;
    try {
      inProgress = g;
      isDL = true;
      
      for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);
	
	debug("loadOnDemand: trial loading "+ m._cn);
	
	try {
	  loadClass(m._cn, true);
	} 
        catch (ClassNotFoundException e) {
	  loaded = false;
	  break;
	}
      }
      
      if (loaded) {
	for (int i = 0; i < g.meths.size(); i++) {
	  Method m = (Method)g.meths.elementAt(i);
	  debug("loadOnDemand: final loading "+ m._cn);
	  
	  try {
	    m.cl = loadClass(m._cn, true);
	  } 
          catch (ClassNotFoundException e) {
	    Entity.error("DLClassLoader: unexpected failure to load "+ m._cn);
	  }
	}
      }
    } 
    finally {      
      // must release
      inProgress = null;
    }
    
    debug("loadOnDemand: end loading "+ g);
    return loaded;
  }
  

  /**
   * Export node and extension classes that capsules may access
   * by adding them to the registry.
   */

  static void exportClass(String name) 
    throws ClassNotFoundException {

    // all exported classes must be locally loadable
    Class cl = Class.forName(name);
    sharedRegistry.put(name, cl);
  }
  
  /** One DLClassLoader is constructed fpr each code group */
  DLClassLoader() {
    inProgress = null;
    isDL = false;
  }
}



 
