package ants;

/** Negotiates a dynamic address on behalf of an ABONE edge node.  This
 * capsule must operate at the so-called virtual-link layer because
 * tunnels and routes have not yet been established
 *
 *@author Andrew Whitaker
 */

public final class DynamicAddressCapsule extends Capsule {
  public static final int JOIN = 1;
  public static final int RESPONSE = 2;
  public static final int LEAVE = 3;
  public static final int ERROR = 4;
  public static final int INIT = 5;
  public static final int UNDEFINED = 503;
  
  public static final int DANTE_VERSION = 1;  
  public static final String VERSION = "ants.DynamicAddressCapsule 1/31/00";
  
  public byte[] mid() { return findMID("ants.DynamicAddressCapsule"); }
  public byte[] gid() { return findGID("ants.DynamicAddressCapsule");}    
  public byte[] pid() { return findPID("ants.DynamicAddressCapsule");}

  protected int type;
  protected ChannelAddress edgeVLLA;
  protected ChannelAddress coreVLLA;
  protected int edgeVNA;
  protected int coreVNA;

  /** We may not have established a gateway or a RouteEntry, 
      so use the raw channel for sending */
  private boolean turnAround (Node n,ChannelAddress uca) {
    int dest = this.getDst();
    this.setDst (this.getSrc());
    this.setSrc (dest);

    return n.link.send (this,uca);
  }

  public boolean evaluate (Node n) {    

    switch (type) {
    case INIT: {
      this.type = JOIN;
      return n.link.send(this,coreVLLA);
    }
    case JOIN: {
      DynamicAddressManager dam = n.getDynamicAddressManager();

      if (dam == null) {
	/* This node is not an active name server.  Sorry. */
	type = ERROR;
	return turnAround(n,edgeVLLA);      
      }

      if (edgeVNA != 0) {
	// this is a keep-alive capsule; add a gateway if necessary
	dam.markAddress(edgeVNA,edgeVLLA);
	return true;
      }
      else {
	int address = dam.getAddress();

	if (address != -1) {
	  type = RESPONSE;
	  edgeVNA = address;
	  coreVNA = n.getAddress();
	
	  return turnAround(n,edgeVLLA);
	}
	else {
	  type = ERROR;
	  return turnAround(n,edgeVLLA);
	}
      }
    }
    case RESPONSE: {
      ChannelAddress damServer = n.getDynamicNameServer();
      if (damServer == null || (! damServer.equals (coreVLLA))) 
	throw new HeaderFormatException ("Unknown name server");
      else {      
	boolean changed = n.setAddress (edgeVNA);
	
	if (changed) {
	  RouteEntry r = new RouteEntry (coreVNA,coreVNA,coreVLLA);

	  // Add a gateway and force a triggered update
	  n.routes.addGateway(coreVNA,r,true);

	  // Send a confirmation to establish a gateway on the server
	  this.type = JOIN;
	  return turnAround(n,coreVLLA);
	}
	else return true;
      }
    }
    case LEAVE: {
      DynamicAddressManager dam = n.getDynamicAddressManager();
      if (dam != null)
	dam.returnAddress(edgeVNA);
      return true;
    }
    case ERROR: {
      Entity.info("DymamicAddressCapsule error");
      return true;
    }
    case UNDEFINED: {
      throw new HeaderFormatException ("Unknown DynamicAddressCapsule type");
    }
    }

    return true;
  }

  public Xdr encode () {
    Xdr xdr = super.encode();
    xdr.PUT ((byte)1); // the version number;
    xdr.PUT ((short) type);
    xdr.PUT ((short) 1); // the number of requested addresses

    /* the refresh timer or the error code; ants ignores this value */
    xdr.PUT ((short) 30); 

    xdr.PUT (edgeVLLA.toString());
    xdr.PUT (coreVLLA.toString());
    
    if (type != ERROR) {
      xdr.PUT (edgeVNA);
      xdr.PUT (coreVNA);
    }
    return xdr;
  }

  public Xdr decode () {
    Xdr xdr = super.decode();
    byte version = xdr.BYTE();
    if ((int) version != DANTE_VERSION)
      throw new HeaderFormatException 
	  ("Unknown DynamicAddressCapsule version");
    

    this.type = (int) xdr.SHORT();
    if (type == UNDEFINED)
	throw new HeaderFormatException 
	    ("Undefined DynamicAddressCapsule type");

    short throwAway = xdr.SHORT();
    throwAway = xdr.SHORT();
    
    try {
      edgeVLLA = ChannelAddress.New(xdr.STRING());
      coreVLLA = ChannelAddress.New(xdr.STRING());
    }
    catch (Exception ex) {
      throw new HeaderFormatException ("Unknown channel address");
    }

    if (type != ERROR) {
      edgeVNA = xdr.INT();
      coreVNA = xdr.INT();
    }
    return xdr;
  }
  
  public int length () {
    int l = super.length() + Xdr.BYTE  + Xdr.SHORT + 
	Xdr.SHORT + Xdr.SHORT + Xdr.STRING(edgeVLLA.toString()) +
	Xdr.STRING (coreVLLA.toString());

    if (type != ERROR)
      l += (2 * Xdr.INT);
    return l;
  }


  public DynamicAddressCapsule (int type,ChannelAddress edgeVLLA,
				ChannelAddress coreVLLA,
				int edgeVNA,int coreVNA) {
    this.edgeVLLA = edgeVLLA;
    this.coreVLLA = coreVLLA;
    this.type = type;
    this.edgeVNA = edgeVNA;
    this.coreVNA = coreVNA;
  }
    
  public DynamicAddressCapsule (int type,ChannelAddress edgeVLLA,
				ChannelAddress coreVLLA) {
    this(type,edgeVLLA,coreVLLA,0,0);
  }
    
  public DynamicAddressCapsule () {
    edgeVNA = 0;
    coreVNA = 0;
    type = UNDEFINED;
  }
    
}
