package ants;

import java.util.*;

/** This class dynamically distributes addresses for a core server.
 * Synchronization is provided by the Noder server object.
 *
 * @author Andrew Whitaker
 */

public final class DynamicAddressManager implements Runnable {
  
  /** The period after which an address is reclaimed.  This value was 
   *  essentially chosen and random; it may need tweeking 
  */
  public static final int TIMEOUT = 300000; // 5 minutes

  /** The frequency of address keep-alives.  Currently, 30 seconds */
  public static final long KEEP_ALIVE_FREQUENCY = (long) (TIMEOUT / 10);

  /** The first address managed by the DAM */
  private int firstAddress;

  /** The number of addresses managed by the DAM */
  private int n;

  /** The number of allocated addresses */
  private int allocated;

  /** A vector of addresses currently in use.  The hash values 
      correspond to boolean use bits */
  private Hashtable addresses;

  /** The local routing table */
  private RouteTable routes;

  /** The first available address (probably) */
  private int index;

  private Node server;

  public DynamicAddressManager 
    (Node server,int firstAddress,int n,RouteTable routes) {

    this.firstAddress = firstAddress;
    this.n = n;
    this.routes = routes;
    this.server = server;

    allocated = 0;
    index = 0;
    addresses = new Hashtable();

    Thread t = new Thread(this);
    t.setDaemon(true);
    t.start();
  }
  
  /** Allocate a new (unused) address.  Assume we are synchronized
      on the Node server */
  public int getAddress () {
    if (allocated == n)
      return -1;

    while (true) {
      Integer key = new Integer (index);
      Boolean b = (Boolean)addresses.get(key);
      if (b == null) {
	allocated++;
	addresses.put(key,Boolean.FALSE);

	int address = firstAddress + index;
	index = (index + 1) % n;
	return address;
      }
      else index = (index + 1) % n;
    }	            
  }

  /** Called on receiving a keep-alive update.  Assume we are 
      synchonized on the Node server */
  public void markAddress (int address,ChannelAddress edgeVLLA) {

    int index = address - firstAddress;
    Integer key = new Integer (index);
    Boolean b = (Boolean) addresses.get(key);
    if (b == Boolean.FALSE) {
      addresses.put(key,Boolean.TRUE);

      // Make sure this gateway is in the table 
      routes.addGateway (address,edgeVLLA);
    }
  }

  /** Return a timed-out address to the pool. Assume we are 
      synchonized on the Node server */
  public void returnAddress(int address) {

    int index = address - firstAddress;
    Integer key = new Integer (index);

    Object o = addresses.get(key);
    if (o != null) {
      allocated--;
      addresses.remove(key);
      
      // Remove this gateway from the routing table
      routes.removeGateway(address);
    }
  }

  /** Clear the use bits after each timeout period */
  public void run () {
    while (true) { 
      try { 
	Thread.sleep(TIMEOUT); 
      } 
      catch (InterruptedException lkjsdf) {}
      
      synchronized(server) {
 
	Enumeration e = addresses.keys();
	while (e.hasMoreElements()) {
	  Integer key = (Integer) e.nextElement();
	  Boolean b = (Boolean)addresses.get(key);

	  if (b == Boolean.TRUE)
	    addresses.put(key,Boolean.FALSE);
	  else if (b == Boolean.FALSE) {
	    returnAddress (key.intValue() + firstAddress);
	  }
	  else {
	    Entity.error("Unknown DynamicAddressManager value");
	  }
	}
      }
    }
  }

}

  

  




