/* 
 *  $Id: Node.java,v 1.4 2000/04/02 21:30:37 andrew Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import java.net.*;
import java.util.*;

import ants.wrapper.*;
import utils.*;


/**
 * node of an active network
 *
 * @author David Wetherall
 */

public class Node extends ManagedObject {
  /** The default number of ABONE addresses that a core node maintains */
  public static final int DEFAULT_ADDRESS_SPACE = 256;

  /** The number of times edge nodes will attempt to acquire an address */
  public static final int RETRY_ATTEMPTS = 5;

  /** The timeout for address request capsules */
  public static final long RETRY_TIMEOUT = 3000; // milleseconds

  public static final String[] defaults = { "-updateRoutes", "false" };

  private boolean running;
  private String routefile;
  private String ANEPfile;
  Capsule context; // accessed by ChannelThreads to set context

  /** The local address of this node */
  private int address;
  public int getAddress() { return address; }

  /** Set the local address if the capsule uniqueID is correct */
  boolean setAddress(int address) { 
    if (this.address != 0)
      return false;
    else {
      synchronized (damServer) {
	this.address =  address;	  
	// Wake up the thread waiting for a dynamic address
	damServer.notify();
	return true;
      }      
    }
  }
      
  /** The default routing table.  Package level access is used to 
   *  allow the routing protocols priveleged access here 
  */
  RouteTable routes;
  RouteTable getRoutes() { return routes; }
  
  private Hashtable alternates;

  void addAlternate(Alternate alt) {
    alternates.put(new Integer(alt.getTypeID()), alt);
  }

  Alternate getAlternate(short typeID) {
    return (Alternate)alternates.get(new Integer(typeID));
  }

  private Hashtable clients;
  Hashtable getApplications() { return clients; }

  /* package level access is used to allow the routing protocols 
     priveleged access here */ 
  Channel link;
  Channel getChannels() { return link; }

  private Hashtable extensions;
  Hashtable getExtensions() { return extensions; }
  
  /* Capsules do have a legitimate reason to access extensions, so the 
   * package-level permission will need to change if extensions are used 
  */
  Extension getExtension(String name) {
    return (Extension) extensions.get(name);
  }

  /** The node soft-store */
  private AutoNodeCache cache;
  public AutoNodeCache getCache() {  return cache;}
  
  /** The capsule code cache; no capsule access here */
  CodeCache capsuleCode;

  /** The dynamic address manager; used by ABONE core nodes */
  DynamicAddressManager dam;
  DynamicAddressManager getDynamicAddressManager() { return dam; }

  /** A dynamic name server.  Used to bootstrap an ABONE edge node */
  ChannelAddress damServer;
  ChannelAddress getDynamicNameServer() { return damServer;}

  /** A thread responsible for sending keep-alives for the dynamic address */
  KeepAliveThread keepAliveThread = null;

  public void setArgs(KeyArgs k) throws Exception {
    // capsules are not allowed to invoke this method
    if (running) return;

    k.merge(defaults);

    int rangeStart = -1;
    int rangeEnd = -1;
    String nameServer = null;

    for (int i = 0; i < k.length(); i++) {
      if (k.key(i).equals("-routes")) {
	routefile = k.arg(i);
	k.strike(i);
      }
      else if (k.key(i).equals("-updateRoutes")) {
	routes.setUpdates(Boolean.valueOf(k.arg(i)).booleanValue());
	k.strike(i);
      }
      else if (k.key(i).equals("-rangeStart")) {
	rangeStart = NodeAddress.fromString(k.arg(i));
	k.strike(i);
      }
      else if (k.key(i).equals("-rangeEnd")) {
	rangeEnd = NodeAddress.fromString(k.arg(i));
	k.strike(i);
      }		  
      else if (k.key(i).equals("-tunnel")) {
	nameServer = k.arg(i);
	k.strike(i);
      }
    }
    
    // initial a dynamic address manager, if needed
    if (rangeStart > 0) {
      int size = rangeEnd - rangeStart;
      
      if (size <= 0) size = DEFAULT_ADDRESS_SPACE;

      Entity.info("Starting dynamic address manager");
      dam = new DynamicAddressManager (this,rangeStart,size,routes);
    }
    else if (nameServer != null) {
      /* We will negotiate an address from this name server once
	 the node has started */
      this.address = 0;
      this.damServer = ChannelAddress.New(nameServer);
    }

    super.setArgs(k);
  }

  public void start() throws Exception {
    // capsules and applications are not allowed to invoke this method
    if (running) return;

    Entity.info("Node " + NodeAddress.toString(address) + " starts");
    if (routefile != null) {
      Entity.info("Reading routes from " + routefile);
      routes.read(routefile);
    }    
    routes.start();

    // export the node environment to capsule classes
    
    Entity.info("exporting classes");
    DLClassLoader.exportClass("ants.Node");
    DLClassLoader.exportClass("ants.Capsule");
    DLClassLoader.exportClass("ants.DataCapsule");
    DLClassLoader.exportClass("ants.Xdr");
    DLClassLoader.exportClass("ants.NodeAddress");
    DLClassLoader.exportClass("ants.ByteArray");
    DLClassLoader.exportClass("ants.wrapper.AutoNodeCache");   
    DLClassLoader.exportClass("ants.NodeCache");
    DLClassLoader.exportClass("ants.Extension");

    DLClassLoader.exportClass("java.lang.Throwable");
    DLClassLoader.exportClass("java.lang.Exception");
    DLClassLoader.exportClass("java.lang.SecurityException");
    DLClassLoader.exportClass("java.lang.ClassNotFoundException");
    DLClassLoader.exportClass("java.lang.Integer");
    DLClassLoader.exportClass("java.lang.Long");
    DLClassLoader.exportClass("java.lang.Object");
    DLClassLoader.exportClass("java.lang.String");
    DLClassLoader.exportClass("java.lang.StringBuffer");
    
    // install the system protocols
    register(new DLProtocol());
    register(new DataProtocol());
    register(new RouteProtocol());

    if ((dam != null) || (damServer != null))
      register(new DynamicAddressProtocol());
  
    // start any node extensions
    Enumeration e2 = extensions.elements();
    while (e2.hasMoreElements())
	((Extension)e2.nextElement()).start();
            
    if (link != null) link.start();

    /** Contact the name server, if necessary */
    if (damServer != null)
      discoverName();

    // now the node is complete
    running = true;

    // start application and network
    Enumeration e = clients.elements();
    while (e.hasMoreElements()) 
      ((Application)e.nextElement()).start();
    
    super.start();
  }

  /** Discover a new ABONE address */
  void discoverName() {
    // start without an address
    address = 0;

    // Stop sending updates for any old address
    if (keepAliveThread != null)
      keepAliveThread.setHalt (true);

    Entity.info ("Contacting address server: " + damServer);

    synchronized (damServer) {
      for (int i=0;i<RETRY_ATTEMPTS;i++) {  
	DynamicAddressCapsule dac = new DynamicAddressCapsule 
	    (DynamicAddressCapsule.INIT,link.source,
	     damServer);
	dac.prime(0,255);      
	
	try {
	  this.context = dac;
	  dac.evaluate(this);
	  damServer.wait(RETRY_TIMEOUT);
	}
	catch (InterruptedException ie) {}
	catch (Exception exc) {
	  Entity.info("Error during name discover: " + exc.toString());
	}
	finally {
	  context = null;
	}
	if (address != 0) {
	  Entity.info ("Acquired dynamic address: " + 
		       NodeAddress.toString(address));

	  /* Start a new thread to send address keep-alives */
	  keepAliveThread = new KeepAliveThread (this);
	  keepAliveThread.setDaemon(true);
	  keepAliveThread.start();
	  
	  return;
	}
      }
    }
    // we couldn't contact the address server
    Entity.error ("Unable to acquire address");
  }
          		  
  /** Register a protocol for use at this node */
  public void register(Protocol p) throws Exception {
    for (int i = 0; i < p.groups.size(); i++) {
      CodeGroup g = (CodeGroup)p.groups.elementAt(i);

      DLClassLoader loader = new DLClassLoader();
      loader.loadForRegister(g);

      if (capsuleCode.retrieve(g.groupID) == null) {
	g.pin = true;
	capsuleCode.set(g);
      }
      
      for (int j = 0; j < g.meths.size(); j++) {
	Method m = (Method)g.meths.elementAt(j);

	if (capsuleCode.lookup(m.methodID) == null) {
	  capsuleCode.enter(m);
	}
      }
    }
  }

  /** Unregister a protocol previously registered at this node */
  public void unregister(Protocol p) {
    // sorry, just an api call for now
  }

  public void attachApplication(Application c, int port) throws Exception {
    Integer key = new Integer(port);

    if (clients.containsKey(key))
      throw new Exception("Port " + port + " is busy.");
    
    clients.put(key, c);
  }
  
  void attachChannel(Channel l) {
    if (link != l) {
      link = l;
      link.attachNode(this);
    }
  }

  void attachExtension(Extension e,String name) {
    extensions.remove (name); // remove any old extension
    extensions.put(name,e);    
  }


  public boolean routeForNode(Capsule c, int dst) {
    RouteEntry r = routes.get(dst);
    
    try {
	return link.send(c, r.addr);
    }
    catch (NullPointerException e) {    
	Entity.info("No route to "+ NodeAddress.toString(dst) 
		    +" from " + NodeAddress.toString(getAddress()));
	throw new NoSuchRouteException();
    }
  }
  
  public boolean deliverToApp(Capsule c, int d) {
    c.saveBuffer = true;
    try {
      ((Application)clients.get(new Integer(d))).receive(c);
      return true;
    }
    catch (NullPointerException e) {
      throw new NoSuchApplicationException();
    }
  }

  /** For use by node components and applications to delay execution */
  public void sleep(long millis) {
    try {
      Thread.sleep(millis);
    } 
    catch (InterruptedException e) {
      ; // who cares
    }
  }
  
  public long time() {
    return System.currentTimeMillis();
  }

  public void log(String msg) {
    Entity.info(this.context.getClass().getName() + " " + 
		NodeAddress.toString(this.getAddress()) + ": " + msg);
  }

  /** Called by Application base class to send a capsule */ 
  synchronized boolean send(Capsule c) {
    try {
      this.context = c;
      return c.evaluate(this);
    }
    finally {
      this.context = null;
    }
  }

  /** constructs a node with given address */
  Node(Manager b, String n, String a) throws Exception {
    super(b, n);

    running = false;
    address = NodeAddress.fromString(a);
    routes = new RouteTable(this);
    cache = new AutoNodeCache(this, 1024, 600);
    capsuleCode = new CodeCache(this, 256);
    extensions = new Hashtable();
    clients = new Hashtable();
    alternates = new Hashtable();

  }

}
