package ants;

import java.net.*;
import java.io.*;

/** This extension of Channel is used to read packets over standard in.
 *  Capsules are still sent over a UDP socket. Anetd requires this facility.
 *
 *  @author Andrew Whitaker
*/

public class StandardInChannel extends Channel {
   public static final int BUFSIZE = 1500;

   protected InputStream inStream;
   protected DatagramSocket socket;

   public int getMTU() { return BUFSIZE; }
   
   public boolean send(byte[] buf, int length, ChannelAddress dst) {
      /* This is the same as UDPChannel.send() */
      
      if (! (dst instanceof UDPChannelAddress)) {
	 Entity.warn("UDPChannel.send() passed non-UDP destination address");
	 return false;
      }
      
      UDPChannelAddress a = (UDPChannelAddress)dst;
      DatagramPacket fr = new DatagramPacket(buf, length, a.address, a.port);

      try {
	 socket.send(fr);
	 if (LOGGING)
	     log(L[7], "transmitted " + fr.getLength() + " bytes");
      } 
      catch (IOException e) {
	 return false;
      }
            
      return true;
   }

   public Capsule receive () throws Exception {
      int avail = 0;
      int bytesRead = 0;
      int totalBytes = 0;
      int temp = 0;

      /* For some reason, this buffer gets reset periodically.  To be on
         the safe side, reset the buffer for each capsule */
      rxbuf = new Xdr (BUFSIZE);
	
      /* Read the packet header to determine the size of the payload */
      while (bytesRead < 8) {

	 try {
	    temp = inStream.read (rxbuf.buf,bytesRead,8-bytesRead);
	 }
	 catch (IOException aba) {
	    /* If we get an exception here, we're in big trouble because the
	       packet header is corrupted.  This means we won't be able to
	       read ANY subsequent packets because we rely on the size field
	       in the packet header to mark capsule boundaries.  Ouch. */
	    
	    if (LOGGING) log (L[1],"inStream.read Exception: " + aba);
	    
	    /* Pass this exception up the chain: it's too important */
	    throw aba;
	 }
	 
	 if (temp > 0) {
	    bytesRead += temp;
	    break;
	 }
	 else Thread.yield();		  
      }
	
      /* The payload size is located at bytes 6 and 7 */
      rxbuf.index = (3 * Xdr.SHORT);
      totalBytes  = rxbuf.SHORT();
      
      if (totalBytes > BUFSIZE)
	  throw new Exception 
	      ("Error: Payload too large for the packet buffer");
      
      
      while (bytesRead < totalBytes) {
	 try {
	    temp = inStream.read(rxbuf.buf,bytesRead,(totalBytes-bytesRead));
	    
	    if (temp > 0) bytesRead += temp;	  	    
	    else Thread.yield();
	 }
	 catch (Exception xxx) {
	    if (LOGGING) log(L[1], "inStream.read Exception: " + xxx);
	 }
      }
      
      if (LOGGING) log(L[7], "received " + bytesRead + " bytes");   
      
      rxbuf.index = 0;
      rxbuf.len = bytesRead;
      return decode(rxbuf); 	
   }
   
   public StandardInChannel(Manager b, String n, UDPChannelAddress src)
     throws IOException {
      
      super(b, n, src);
      
      socket = new DatagramSocket();
      inStream = System.in;
      
      txbuf = new Xdr(BUFSIZE);
      rxbuf = new Xdr(BUFSIZE);
   }


   public StandardInChannel(Manager b, String n, String hp)
     throws Exception {
      
      this(b, n, new UDPChannelAddress(hp));
   }
}
	  
	   
   
      
